// Tonelli-Shanks Algorithm. (Ice Cracker code)
//Code from 'Geeks for Geeks'. (chinmoy1997pal)
//Given two values n and a prime number p. Find Square Root under Modulo p.
//Note original 'H.Cohen' code. page 32. 1.5.1 Requires (a/p) = -1. Quadratic Non Reciprocity.
//"Thus a direct use would require factoring all the numbers into prime"
//cont.. "The formula (Theorem 1.4.7) are still true if p and q are only supposed to be
//positive odd integers". Ie not primes
//
//See my "Shankers" C++ Code download for non random 'n' example.(provisional)
//Great introduction code for such as Schoofs, due to the bounds required for large 'n'

#include <iostream>
#include <math.h>
using namespace std;
int powerMod(int base, int exponent, int modulus) {  //Choose number n at random until (n/p) = -1. Set z = n^q(mod p)
    int result = 1;                                  //function finds pow(base, exponent) % modulus
    base = base % modulus;
    while (exponent > 0) {
        if (exponent % 2 == 1)
            result = (result * base) % modulus;
        exponent = exponent >> 1;                    //shift bit right, equiv divides by 2(1*n)
        base = (base * base) % modulus;
    }
    return result;
}
int gcd(int a, int b) {                              //function finds GCD (co-prime)
    if (b == 0)
        return a;
    else
        return gcd(b, a % b);
}
int orderValues(int p, int b) {                     //function returns K for ( b^k = 1(mod p) )
    if (gcd(p, b) != 1) {
        cout << "p and b are not Co-prime" << endl;
        return -1;
    }
    int k = 3;                                      // starts 'k' with first odd prime
    while (1) {
        if (powerMod(b, k, p) == 1)
            return k;
        k++;
    }
}
int findx2e(int x, int& e) {                        // returns p - 1 (x = arg) as x * 2^e
    e = 0;                                        // int& e is a "move semantic 'L Value' ongoing value (function calls that return a reference are lvalues)
    while (x % 2 == 0) {                            // x is odd
        x /= 2;
        e++;
        //cout << "e = " << e << endl;
    }
    return x;
}
int calcSquareRoot(int n, int p) {                   //Input Start
    if (gcd(n, p) != 1) {
        cout << "p and n are not Co-prime" << endl;
        return -1;
    }
    if (powerMod(n, (p - 1) / 2, p) == (p - 1)) {
        cout << "square root not possible if return is (p -1)" << endl;
        return -1;
    }
    int s, e;                                         // P - 1, in terms of s * 2^e, s being odd.
    s = findx2e(p - 1, e);
    int q;                                            // finding smallest q
    for (q = 2; ; q++) {
        if (powerMod(q, (p - 1) / 2, p) == (p - 1))
            //cout << "q = " << q << endl;
            break;
    }
    int x = powerMod(n, (s + 1) / 2, p);              // Initialise
    int b = powerMod(n, s, p);
    int g = powerMod(q, s, p);
    int r = e;
    while (1) {                                       // loop determines wether b == 1 or m becomes 0
        int m;
        for (m = 0; m < r; m++) {
            if (orderValues(p, b) == -1)
                return -1;
            if (orderValues(p, b) == pow(2, m))       // m such that b^(2^m) = 1
                break;
        }
        if (m == 0)
            return x;
        x = (x * powerMod(g, pow(2, r - m - 1), p)) % p;  // Updates 
        g = powerMod(g, pow(2, r - m), p);
        b = (b * g) % p;
        if (b == 1)
            return x;
        r = m;
    }
}
int main() {
    int n = 4;
    int p = 11;
    int sqrtVal = calcSquareRoot(n, p);
    if (sqrtVal == -1)
        cout << "Modular square root is not exist";
    else
        cout << "Modular square root of the number is " << sqrtVal;
}